/*
 * Copyright 2012-2013 Canoo Engineering AG.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opendolphin.core.client.comm

import groovyx.gpars.dataflow.KanbanFlow
import groovyx.gpars.dataflow.KanbanTray
import groovyx.gpars.dataflow.ProcessingNode
import org.opendolphin.core.Attribute
import org.opendolphin.core.PresentationModel
import org.opendolphin.core.client.ClientAttribute
import org.opendolphin.core.client.ClientDolphin
import org.opendolphin.core.client.ClientModelStore
import org.opendolphin.core.client.ClientPresentationModel
import org.opendolphin.core.comm.*
import groovy.transform.CompileStatic
import groovy.util.logging.Log
import org.codehaus.groovy.runtime.StackTraceUtils

import java.beans.PropertyChangeEvent
import java.beans.PropertyChangeListener
import java.util.logging.Level

@Log
abstract class ClientConnector {
    Codec codec
    UiThreadHandler uiThreadHandler // must be set from the outside - toolkit specific

    Closure onException = { Throwable up ->
        def out = new StringWriter()
        up.printStackTrace(new PrintWriter(out))
        log.severe("onException reached, rethrowing in UI Thread, consider setting ClientConnector.onException\n${out.buffer}")
        uiThreadHandler.executeInsideUiThread { throw up } // not sure whether this is a good default
    }

    protected final ClientDolphin clientDolphin
    protected final CommandBatcher commandBatcher

    ClientConnector(ClientDolphin clientDolphin) {
        this(clientDolphin, null)
    }

    ClientConnector(ClientDolphin clientDolphin, CommandBatcher commandBatcher) {
        this.clientDolphin = clientDolphin
        this.commandBatcher = commandBatcher ?: new CommandBatcher()

        startCommandProcessing()

    }

    protected void startCommandProcessing() {

        def transmitter = ProcessingNode.node { trayOut ->
            def commandsAndHandlers = commandBatcher.waitingBatches.val
            List<Command> commands = commandsAndHandlers.collect { it.command }
            if (log.isLoggable(Level.INFO)) {
                log.info "C: sending batch of size " + commands.size()
                for (command in commands) { log.info("C:           -> "+command)}
            }

            def answer = null
            doExceptionSafe(
                { answer = transmit(commands) },
                { trayOut << [response: answer, request: commandsAndHandlers] })
        }

        def worker = ProcessingNode.node { KanbanTray trayIn ->
            Map got = trayIn.take()
            if (got.response == null) return // we cannot ignore empty responses. They may have an onFinished handler
            insideUiThread {
                processResults(got.response, got.request)
            }
        }

        KanbanFlow flow = new KanbanFlow()
        flow.link transmitter to worker
        flow.start(1)
    }

    protected getClientModelStore() {
        clientDolphin.clientModelStore
    }

    abstract List<Command> transmit(List<Command> commands)

    @CompileStatic
    void send(Command command, OnFinishedHandler callback = null) {
        // we are inside the UI thread and events calls come in strict order as received by the UI toolkit
        commandBatcher.batch(new CommandAndHandler(command: command, handler: callback))
    }

    @CompileStatic
    void processResults(List<Command> response, List<CommandAndHandler> commandsAndHandlers) {
        def me = this
        me.info "C: server responded with ${response?.size()} command(s): ${response?.id}"

        List<ClientPresentationModel> touchedPresentationModels = new LinkedList<ClientPresentationModel>()
        List<Map> touchedDataMaps = new LinkedList<Map>()
        for (Command serverCommand in response) {
            def touched = me.dispatchHandle serverCommand
            if (touched && touched instanceof ClientPresentationModel) {
                touchedPresentationModels << (ClientPresentationModel) touched
            } else if (touched && touched instanceof Map) {
                touchedDataMaps << (Map) touched
            }
        }
        def callback = commandsAndHandlers.first().handler // there can only be one relevant handler anyway
        if (callback) {
            callback.onFinished((List<ClientPresentationModel>) touchedPresentationModels.unique { ((ClientPresentationModel) it).id })
            callback.onFinishedData(touchedDataMaps)
        }
    }

    void info(Object message) {
        log.info message
    }

    Object dispatchHandle(Command command) {
        handle command
    }

    @CompileStatic
    void doExceptionSafe(Runnable processing, Runnable atLeast = null) {
        try {
            processing.run()
        } catch (e) {
            StackTraceUtils.deepSanitize(e)
            onException e
        } finally {
            if (atLeast) atLeast.run()
        }
    }

    @CompileStatic
    void insideUiThread(Runnable processing, Runnable atLeast = null) {
        doExceptionSafe ({
            if (uiThreadHandler) {
                uiThreadHandler.executeInsideUiThread(processing)
            } else {
                println("please provide howToProcessInsideUI handler")
                processing.run()
            }
        }, atLeast)
    }

    def handle(Command serverCommand) {
        log.severe "C: cannot handle unknown command '$serverCommand'"
    }

    Map handle(DataCommand serverCommand) {
        return serverCommand.data
    }

    ClientPresentationModel handle(DeletePresentationModelCommand serverCommand) {
        ClientPresentationModel model = clientDolphin.findPresentationModelById(serverCommand.pmId)
        if (!model) return null
        clientModelStore.delete(model)
        return model
    }

    ClientPresentationModel handle(DeleteAllPresentationModelsOfTypeCommand serverCommand) {
        clientDolphin.deleteAllPresentationModelsOfType(serverCommand.pmType)
        return null // we cannot really return a single pm here
    }

    @CompileStatic
    ClientPresentationModel handle(CreatePresentationModelCommand serverCommand) {
        if (((ClientModelStore) clientModelStore).containsPresentationModel(serverCommand.pmId)) {
            throw new IllegalStateException("There already is a presentation model with id '$serverCommand.pmId' known to the client.")
        }
        List<ClientAttribute> attributes = []
        for (attr in serverCommand.attributes) {
            ClientAttribute attribute = new ClientAttribute(attr.propertyName.toString(), attr.value)
            attribute.qualifier = attr.qualifier
            attributes << attribute
        }
        ClientPresentationModel model = new ClientPresentationModel(serverCommand.pmId, attributes)
        model.presentationModelType = serverCommand.pmType
        if (serverCommand.clientSideOnly) {
            model.clientSideOnly = true
            ((ClientModelStore)clientModelStore).addClientSideOnly(model)
        } else {
            ((ClientModelStore)clientModelStore).add(model)
        }
        clientDolphin.updateQualifiers(model)
        return model
    }

    ClientPresentationModel handle(ValueChangedCommand serverCommand) {
        Attribute attribute = clientModelStore.findAttributeById(serverCommand.attributeId)
        if (!attribute) {
            log.warning "C: attribute with id '$serverCommand.attributeId' not found, cannot update old value '$serverCommand.oldValue' to new value '$serverCommand.newValue'"
            return null
        }
        log.info "C: updating '$attribute.propertyName' id '$serverCommand.attributeId' from '$attribute.value' to '$serverCommand.newValue'"
        attribute.value = serverCommand.newValue
        return null // this command is not expected to be sent explicitly, so no pm needs to be returned
    }

    ClientPresentationModel handle(BaseValueChangedCommand serverCommand) {
        Attribute attribute = clientModelStore.findAttributeById(serverCommand.attributeId)
        if (!attribute) {
            log.warning "C: attribute with id '$serverCommand.attributeId' not found, cannot set initial value."
            return null
        }
        log.info "C: updating id '$serverCommand.attributeId' setting initialValue to '$attribute.value'"
        attribute.rebase()
        return null // this command is not expected to be sent explicitly, so no pm needs to be returned
    }

    ClientPresentationModel handle(SwitchPresentationModelCommand serverCommand) {
        def switchPm = clientModelStore.findPresentationModelById(serverCommand.pmId)
        if (!switchPm) {
            log.warning "C: switch pm with id '$serverCommand.pmId' not found, cannot switch"
            return null
        }
        def sourcePm = clientModelStore.findPresentationModelById(serverCommand.sourcePmId)
        if (!sourcePm) {
            log.warning "C: source pm with id '$serverCommand.sourcePmId' not found, cannot switch"
            return null
        }
        switchPm.syncWith sourcePm                  // ==  clientDolphin.apply sourcePm to switchPm
        return (ClientPresentationModel) switchPm
    }

    ClientPresentationModel handle(InitializeAttributeCommand serverCommand) {
        def attribute = new ClientAttribute(serverCommand.propertyName, serverCommand.newValue, serverCommand.qualifier, serverCommand.tag)

        // todo: add check for no-value; null is a valid value
        if (serverCommand.qualifier) {
            def copies = clientModelStore.findAllAttributesByQualifier(serverCommand.qualifier)
            if (copies) {
                if (null == serverCommand.newValue) {
                    attribute.value = copies.first()?.value
                } else {
                    copies.each { attr ->
                        attr.value = attribute.value
                    }
                }
            }
        }
        def presentationModel = null
        if (serverCommand.pmId) presentationModel = clientModelStore.findPresentationModelById(serverCommand.pmId)
        // if there is no pmId, it is most likely an error and CreatePresentationModelCommand should have been used
        if (!presentationModel) {
            presentationModel = new ClientPresentationModel(serverCommand.pmId, [])
            presentationModel.setPresentationModelType(serverCommand.pmType)
            clientModelStore.add(presentationModel)
        }
        clientDolphin.addAttributeToModel(presentationModel, attribute)
        clientDolphin.updateQualifiers(presentationModel)
        return presentationModel // todo dk: check and test
    }

    ClientPresentationModel handle(SavedPresentationModelNotification serverCommand) {
        if (!serverCommand.pmId) return null
        ClientPresentationModel model = clientModelStore.findPresentationModelById(serverCommand.pmId)
        if (null == model) {
            log.warning("model with id '$serverCommand.pmId' not found, cannot rebase")
            return null
        }
        model.attributes*.rebase() // rebase sends update command if needed through PCL
        return model
    }

    ClientPresentationModel handle(PresentationModelResetedCommand serverCommand) {
        if (!serverCommand.pmId) return null
        PresentationModel model = clientModelStore.findPresentationModelById(serverCommand.pmId)
        // reset locally first
        if (!model) return null
        model.attributes*.reset()
        return model
    }

    ClientPresentationModel handle(AttributeMetadataChangedCommand serverCommand) {
        ClientAttribute attribute = clientModelStore.findAttributeById(serverCommand.attributeId)
        if (!attribute) return null
        attribute[serverCommand.metadataName] = serverCommand.value
        return null
    }

}